from typing import Dict, List

from duckdb import DuckDBPyConnection
from graphviz import Digraph

from . import Constraint, Column, ForeignKey, Table


class VizDrawer:
    def __init__(self, con: DuckDBPyConnection):
        self.tables: List[Table] = []

        tables: Dict[str, Table] = {}
        constraints: Dict[int, Constraint] = {}

        # Get table names first. In the columns table we can not filter
        # for base tables and some of the tables might not be contained
        # in the constraints' information.
        for table_name, in con.execute('''
            SELECT table_name
            FROM information_schema.tables
            WHERE table_type == 'BASE TABLE'
        ''').fetchall():
            table = Table(table_name)

            self.tables.append(table)
            tables[table_name] = table

        # Get column names and data types for each table.
        for table_name, column_name, data_type in con.execute('''
            SELECT
                table_name,
                column_name,
                data_type
            FROM information_schema.columns
            ORDER BY ordinal_position ASC
        ''').fetchall():
            if table_name in tables:
                table = tables[table_name]

                column = Column(table, column_name, data_type)
                table.columns.append(column)

        # Find primary keys.
        for table_name, constraint_index, constraint_columns in con.execute('''
            SELECT
                table_name,
                constraint_index,
                constraint_column_names
            FROM duckdb_constraints()
            WHERE constraint_type = 'PRIMARY KEY'
            ORDER BY constraint_index ASC
        ''').fetchall():
            # get table
            if table_name not in tables:
                raise AssertionError(f'unknown table {table_name} for constraint {constraint_index}')

            table = tables[table_name]

            # store constraint
            if constraint_index in constraints:
                raise AssertionError(f'constraint with index {constraint_index} already stored')

            constraint = Constraint(
                constraint_index,
                table,
                tuple(table.get_column(c) for c in constraint_columns)
            )
            constraints[constraint_index] = constraint

            # store key
            if table.primary_key is not None:
                raise AssertionError(f'discovered second primary key for table {table_name}')

            table.primary_key = constraint

        # Find unique keys.
        for table_name, constraint_index, constraint_columns in con.execute('''
            SELECT
                table_name,
                constraint_index,
                constraint_column_names
            FROM duckdb_constraints()
            WHERE constraint_type = 'UNIQUE'
            ORDER BY constraint_index ASC
        ''').fetchall():
            # get table
            if table_name not in tables:
                raise AssertionError(f'unknown table {table_name} for constraint {constraint_index}')

            table = tables[table_name]

            # store constraint
            if constraint_index in constraints:
                raise AssertionError(f'constraint with index {constraint_index} already stored')

            constraint = Constraint(
                constraint_index,
                table,
                tuple(table.get_column(c) for c in constraint_columns)
            )
            constraints[constraint_index] = constraint

            # store key
            table.unique_keys.append(constraint)

        # Find foreign keys.
        for table_name, constraint_index, constraint_columns in con.execute('''
            SELECT
                table_name,
                constraint_index,
                constraint_column_names
            FROM duckdb_constraints()
            WHERE constraint_type = 'FOREIGN KEY'
            ORDER BY constraint_index ASC
        ''').fetchall():
            # get table
            if table_name not in tables:
                raise AssertionError(f'unknown table {table_name} for constraint {constraint_index}')

            table = tables[table_name]

            # lookup constraint
            if constraint_index not in constraints:
                raise AssertionError(f'constraint with index {constraint_index} not discovered previously')

            constraint = constraints[constraint_index]

            # store key
            key = ForeignKey(tuple(table.get_column(c) for c in constraint_columns), constraint)
            table.foreign_keys.append(key)

    def to_graph(self) -> Digraph:
        # create graph
        ps = Digraph('Schema',
                     graph_attr={},
                     node_attr={
                         'shape': 'plaintext'
                     })

        # add nodes
        fk_counter: Dict[str, int] = {}

        for table in self.tables:
            columns = "\n".join(self.__column_to_html(table, column, fk_counter) for column in table.columns)

            ps.node(
                table.id,
                f'''<
                    <table border="0" cellborder="1" cellspacing="0" cellpadding="5">
                        <tr>
                            <td><b>{table.name}</b></td>
                        </tr>
                        <tr>
                            <td>
                                <table border="0" cellborder="0" cellspacing="0">
                                    {columns}
                                </table>
                            </td>
                        </tr>
                    </table>
                >'''
            )

        # add edges
        for source_table in self.tables:
            for key in source_table.foreign_keys:
                target_table = key.constraint.table
                fk_counter_key = f'{source_table.name}_{key.constraint.index}'

                ps.edge(source_table.id, target_table.id, label=f'FK{fk_counter[fk_counter_key]}', arrowhead='vee')

        # return graph
        return ps

    def to_svg(self, lr: bool) -> str:
        ps = self.to_graph()
        if lr:
            ps.graph_attr['rankdir'] = 'LR'

        return ps.pipe(format='svg').decode('utf-8')

    @staticmethod
    def __column_to_html(table: Table, column: Column, fk_counter: Dict[str, int]):
        name = column.name
        data_type = column.data_type

        # extract and style column name
        if table.primary_key is not None and column in table.primary_key.columns:
            name = f'<b>{name}</b>'
        for key in table.unique_keys:
            if column in key.columns:
                name = f'<u>{name}</u>'
                break

        # extract foreign keys
        fk = []
        for key in table.foreign_keys:
            if column in key.columns:
                fk_counter_key = f'{table.name}_{key.constraint.index}'
                if fk_counter_key not in fk_counter:
                    fk_counter[fk_counter_key] = max(*fk_counter.values(), 0, 0) + 1

                fk.append(fk_counter[fk_counter_key])

        if len(fk) > 0:
            fk = map(lambda x: f'(FK{x})', sorted(fk))
            fk = f'<i>{" ".join(fk)}</i>'
        else:
            fk = ''

        # convert to html
        return f'''
            <tr port="{column.id}">
                <td align="left">{name}</td>
                <td align="left">: {data_type}</td>
                <td align="left">{fk}</td>
            </tr>
        '''
